from .pdg_format import _round, _strip
import numpy as np
import re


def pdg_format(value, *errors):
    if value is None:
        strings, nexp = _round((0, *errors), None, None)
        strings = strings[1:]
    else:
        strings, nexp = _round((value, *errors), None, None)
    strings = _strip(strings)
    if nexp == 0:
        return strings
    for i, s in enumerate(strings):
        m = re.match(r"(-?)0\.0+", s)
        s = m.group(1) + "0" if m else "%sE%i" % (s, nexp)
        strings[i] = s
    return strings


def matrix_format(*values):
    vs = np.array(values)
    mv = np.max(np.abs(vs))
    smv = "%.3G" % mv
    try:
        i = smv.index("E")
        sexp = smv[i + 1 :]
        exp = int(sexp)
        vs /= 10 ** exp
        s = ["%.3fE%s" % (v, sexp) for v in vs]
    except ValueError:
        s = ["%.3f" % v for v in vs]
    return _strip(s)


def goaledm(fm):
    # taken from the source code, see VariableMeticBuilder.cxx
    return 2e-3 * fm.tolerance * fm.up


def format_row(widths, *args):
    return (
        "".join(
            ("|{0:^%i}" % (w - 1) if w > 0 else "| {0:%i}" % (-w - 2)).format(a)
            for (w, a) in zip(widths, args)
        )
        + "|"
    )


def fmin(fm):
    ws = (-32, 33)
    i1 = format_row(
        ws, "FCN = %.4G" % fm.fval, "Ncalls=%i (%i total)" % (fm.nfcn, fm.ncalls)
    )
    i2 = format_row(
        ws, "EDM = %.3G (Goal: %G)" % (fm.edm, goaledm(fm)), "up = %.1f" % fm.up
    )
    ws = (16, 16, 12, 21)
    h1 = format_row(ws, "Valid Min.", "Valid Param.", "Above EDM", "Reached call limit")
    v1 = format_row(
        ws,
        repr(fm.is_valid),
        repr(fm.has_valid_parameters),
        repr(fm.is_above_max_edm),
        repr(fm.has_reached_call_limit),
    )
    ws = (16, 16, 12, 12, 9)
    h2 = format_row(ws, "Hesse failed", "Has cov.", "Accurate", "Pos. def.", "Forced")
    v2 = format_row(
        ws,
        repr(fm.hesse_failed),
        repr(fm.has_covariance),
        repr(fm.has_accurate_covar),
        repr(fm.has_posdef_covar),
        repr(fm.has_made_posdef_covar),
    )

    ln = len(h1) * "-"
    return "\n".join((ln, i1, i2, ln, h1, ln, v1, ln, h2, ln, v2, ln))


def params(mps):
    vnames = [mp.name for mp in mps]
    name_width = max([4] + [len(x) for x in vnames])
    num_width = max(2, len("%i" % (len(vnames) - 1)))

    ws = (-num_width - 2, -name_width - 3, 12, 12, 13, 13, 10, 10, 8)
    h = format_row(
        ws,
        "",
        "Name",
        "Value",
        "Hesse Err",
        "Minos Err-",
        "Minos Err+",
        "Limit-",
        "Limit+",
        "Fixed",
    )
    ln = "-" * len(h)
    lines = [ln, h, ln]
    mes = mps.merrors
    for i, mp in enumerate(mps):
        if mes and mp.name in mes:
            me = mes[mp.name]
            val, err, mel, meu = pdg_format(mp.value, mp.error, me.lower, me.upper)
        else:
            val, err = pdg_format(mp.value, mp.error)
            mel = ""
            meu = ""
        lines.append(
            format_row(
                ws,
                str(i),
                mp.name,
                val,
                err,
                mel,
                meu,
                "%g" % mp.lower_limit if mp.lower_limit is not None else "",
                "%g" % mp.upper_limit if mp.upper_limit is not None else "",
                "yes" if mp.is_fixed else "CONST" if mp.is_const else "",
            )
        )
    lines.append(ln)
    return "\n".join(lines)


def merror(me):
    mel, meu = pdg_format(None, me.lower, me.upper)
    stat = "Valid" if me.is_valid else "Invalid"
    summary = "| {0:^15s} | {1:^27s} |".format(me.name, stat)
    error = "| {0:^15s} | {1:^12s} | {2:^12s} |".format("Error", mel, meu)
    valid = "| {0:^15s} | {1:^12s} | {2:^12s} |".format(
        "Valid", str(me.lower_valid), str(me.upper_valid)
    )
    at_limit = "| {0:^15s} | {1:^12s} | {2:^12s} |".format(
        "At Limit", str(me.at_lower_limit), str(me.at_upper_limit)
    )
    max_fcn = "| {0:^15s} | {1:^12s} | {2:^12s} |".format(
        "Max FCN", str(me.at_lower_max_fcn), str(me.at_upper_max_fcn)
    )
    new_min = "| {0:^15s} | {1:^12s} | {2:^12s} |".format(
        "New Min", str(me.lower_new_min), str(me.upper_new_min)
    )
    hline = "-" * len(error)
    return "\n".join(
        (hline, summary, hline, error, valid, at_limit, max_fcn, new_min, hline)
    )


def matrix(m):
    n = len(m)

    args = []
    for mi in m:
        for mj in mi:
            args.append(mj)
    nums = matrix_format(*args)

    def row_fmt(args):
        s = "| " + args[0] + " |"
        for x in args[1:]:
            s += " " + x
        s += " |"
        return s

    first_row_width = max(len(v) for v in m.names)
    row_width = max(first_row_width, max(len(v) for v in nums))
    v_names = [("{:>%is}" % first_row_width).format(x) for x in m.names]
    h_names = [("{:>%is}" % row_width).format(x) for x in m.names]
    val_fmt = ("{:>%is}" % row_width).format

    header = row_fmt([" " * first_row_width] + h_names)
    hline = "-" * len(header)
    lines = [hline, header, hline]

    for i, vn in enumerate(v_names):
        lines.append(row_fmt([vn] + [val_fmt(nums[n * i + j]) for j in range(n)]))
    lines.append(hline)
    return "\n".join(lines)



import pytest
from pytest import approx
from iminuit import Minuit
from sys import version_info as pyver


class Fcn:
    def __call__(self, x):
        return x ** 2

    def default_errordef(self):
        return 4


@pytest.fixture
def minuit():
    m = Minuit(lambda x: x ** 2, pedantic=False)
    m.migrad()
    return m


def test_default_errordef():
    with pytest.warns(DeprecationWarning):
        Minuit(Fcn(), pedantic=False)


def test_edm(minuit):
    with pytest.warns(DeprecationWarning):
        assert minuit.edm == approx(0)


def test_matrix_accurate(minuit):
    with pytest.warns(DeprecationWarning):
        assert minuit.matrix_accurate() == minuit.accurate


def test_get_fmin(minuit):
    with pytest.warns(DeprecationWarning):
        assert minuit.get_fmin() == minuit.fmin


def test_get_param_states(minuit):
    with pytest.warns(DeprecationWarning):
        assert minuit.get_param_states() == minuit.params


def test_get_initial_param_states(minuit):
    with pytest.warns(DeprecationWarning):
        assert minuit.get_initial_param_states() == minuit.init_params


def test_get_num_call_fcn(minuit):
    with pytest.warns(DeprecationWarning):
        assert minuit.get_num_call_fcn() == minuit.ncalls_total


def test_num_call_grad(minuit):
    with pytest.warns(DeprecationWarning):
        assert minuit.get_num_call_grad() == minuit.ngrads_total


def test_is_fixed(minuit):
    with pytest.warns(DeprecationWarning):
        assert minuit.is_fixed("x") is False

    minuit.fixed["x"] = True
    with pytest.warns(DeprecationWarning):
        assert minuit.is_fixed("x") is True
    minuit.fixed["x"] = False


def test_migrad_ok(minuit):
    with pytest.warns(DeprecationWarning):
        assert minuit.migrad_ok() == minuit.valid


def test_print_matrix(minuit):
    minuit.hesse()
    with pytest.warns(DeprecationWarning):
        minuit.print_matrix()


def test_print_fmin(minuit):
    with pytest.warns(DeprecationWarning):
        minuit.print_fmin()


def test_print_all_minos(minuit):
    with pytest.warns(DeprecationWarning):
        minuit.print_all_minos()


def test_print_param(minuit):
    with pytest.warns(DeprecationWarning):
        minuit.print_param()


def test_print_initial_param(minuit):
    with pytest.warns(DeprecationWarning):
        minuit.print_initial_param()


def test_set_errordef(minuit):
    assert minuit.errordef == 1
    with pytest.warns(DeprecationWarning):
        minuit.set_errordef(4)
    assert minuit.errordef == 4


def test_set_up(minuit):
    assert minuit.errordef == 1
    with pytest.warns(DeprecationWarning):
        minuit.set_up(4)
    assert minuit.errordef == 4


def test_set_strategy(minuit):
    assert minuit.strategy == 1
    with pytest.warns(DeprecationWarning):
        minuit.set_strategy(2)
    assert minuit.strategy == 2


def test_set_print_level(minuit):
    minuit.print_level = 0
    with pytest.warns(DeprecationWarning):
        minuit.set_print_level(2)
    assert minuit.print_level == 2


def test_hesse_maxcall(minuit):
    with pytest.warns(DeprecationWarning):
        minuit.hesse(maxcall=10)


def test_list_of_vary_param(minuit):
    with pytest.warns(DeprecationWarning):
        assert minuit.list_of_vary_param() == [
            k for (k, v) in minuit.fixed.items() if not v
        ]


def test_list_of_fixed_param(minuit):
    with pytest.warns(DeprecationWarning):
        assert minuit.list_of_fixed_param() == [
            k for (k, v) in minuit.fixed.items() if v
        ]


def test_import_iminuit_warnings():
    with pytest.warns(DeprecationWarning):
        import iminuit.iminuit_warnings  # noqa: F401


@pytest.mark.skipif(
    pyver < (3, 7),
    reason="Deprecating module-level objects requires python-3.7 or newer",
)
def test_import_from_iminuit_warnings():
    with pytest.warns(DeprecationWarning):
        from iminuit.iminuit_warnings import IMinuitWarning  # noqa: F401



# flake8: noqa E501
from iminuit import Minuit
from iminuit.util import Params, Param, Matrix
from iminuit import repr_html, repr_text
import pytest


def f1(x, y):
    return (x - 2) ** 2 + (y - 1) ** 2 / 0.25 + 1


def test_html_tag():
    s = repr_html.Html()
    with repr_html.Tag("foo", s):
        s += "bar"
    assert str(s) == "<foo>\nbar\n</foo>\n"


def test_pdg_format():
    assert repr_text.pdg_format(1.2567, 0.1234) == ["1.26", "0.12"]
    assert repr_text.pdg_format(1.2567e3, 0.1234e3) == ["1.26E3", "0.12E3"]
    assert repr_text.pdg_format(1.2567e4, 0.1234e4) == ["12.6E3", "1.2E3"]
    assert repr_text.pdg_format(1.2567e-1, 0.1234e-1) == ["0.126", "0.012"]
    assert repr_text.pdg_format(1.2567e-2, 0.1234e-2) == ["12.6E-3", "1.2E-3"]
    assert repr_text.pdg_format(1.0, 0.0, 0.25) == ["1.00", "0.00", "0.25"]
    assert repr_text.pdg_format(0, 1, -1) == ["0", "1", "-1"]
    assert repr_text.pdg_format(2, -1, 1) == ["2", "-1", "1"]
    assert repr_text.pdg_format(2.01, -1.01, 1.01) == ["2", "-1", "1"]
    assert repr_text.pdg_format(1.999, -0.999, 0.999) == ["2", "-1", "1"]
    assert repr_text.pdg_format(1, 0.5, -0.5) == ["1.0", "0.5", "-0.5"]
    assert repr_text.pdg_format(1.0, 1e-4) == ["1000.0E-3", "0.1E-3"]
    assert repr_text.pdg_format(-1.234567e-22, 1.234567e-11) == ["-0", "1.2E-11"]
    nan = float("nan")
    assert repr_text.pdg_format(nan, 1.23e-2) == ["nan", "0.012"]
    assert repr_text.pdg_format(nan, -nan) == ["nan", "nan"]


def test_matrix_format():
    nan = float("nan")
    assert repr_text.matrix_format(1e1, 2e2, -3e3, -4e4) == [
        "0.001E+04",
        "0.020E+04",
        "-0.300E+04",
        "-4.000E+04",
    ]
    assert repr_text.matrix_format(nan, 2e2, -nan, -4e4) == [
        "nan",
        "200",
        "nan",
        "-40000",
    ]


@pytest.fixture
def minuit():
    m = Minuit(f1, x=0, y=0, pedantic=False, print_level=0)
    m.tol = 1e-4
    m.migrad()
    m.hesse()
    m.minos()
    return m


def test_html_fmin(minuit):
    fmin = minuit.fmin
    assert (
        fmin._repr_html_()
        == r"""<table>
<tr>
<td colspan="2" title="Minimum value of function">
FCN = 1
</td>
<td align="center" colspan="3" title="No. of calls in last algorithm and total number of calls">
Ncalls = 34 (58 total)
</td>
</tr>
<tr>
<td colspan="2" title="Estimated distance to minimum and target threshold">
EDM = %.3G (Goal: 2E-07)
</td>
<td align="center" colspan="3" title="Increase in FCN which corresponds to 1 standard deviation">
up = 1.0
</td>
</tr>
<tr>
<td align="center" title="Validity of the migrad call">
Valid Min.
</td>
<td align="center" title="Validity of parameters">
Valid Param.
</td>
<td align="center" title="Is EDM above goal EDM?">
Above EDM
</td>
<td align="center" colspan="2" title="Did last migrad call reach max call limit?">
Reached call limit
</td>
</tr>
<tr>
<td align="center" style="background-color:#92CCA6;">
True
</td>
<td align="center" style="background-color:#92CCA6;">
True
</td>
<td align="center" style="background-color:#92CCA6;">
False
</td>
<td align="center" colspan="2" style="background-color:#92CCA6;">
False
</td>
</tr>
<tr>
<td align="center" title="Did Hesse fail?">
Hesse failed
</td>
<td align="center" title="Has covariance matrix">
Has cov.
</td>
<td align="center" title="Is covariance matrix accurate?">
Accurate
</td>
<td align="center" title="Is covariance matrix positive definite?">
Pos. def.
</td>
<td align="center" title="Was positive definiteness enforced by Minuit?">
Forced
</td>
</tr>
<tr>
<td align="center" style="background-color:#92CCA6;">
False
</td>
<td align="center" style="background-color:#92CCA6;">
True
</td>
<td align="center" style="background-color:#92CCA6;">
True
</td>
<td align="center" style="background-color:#92CCA6;">
True
</td>
<td align="center" style="background-color:#92CCA6;">
False
</td>
</tr>
</table>
"""
        % fmin.edm
    )


def test_html_params(minuit):
    assert (
        r"""<table>
<tr style="background-color:#F4F4F4;">
<td/>
<th title="Variable name">
Name
</th>
<th title="Value of parameter">
Value
</th>
<th title="Hesse error">
Hesse Error
</th>
<th title="Minos lower error">
Minos Error-
</th>
<th title="Minos upper error">
Minos Error+
</th>
<th title="Lower limit of the parameter">
Limit-
</th>
<th title="Upper limit of the parameter">
Limit+
</th>
<th title="Is the parameter fixed in the fit">
Fixed
</th>
</tr>
<tr style="background-color:#FFFFFF;">
<td>
0
</td>
<td>
x
</td>
<td>
0
</td>
<td>
1
</td>
<td>

</td>
<td>

</td>
<td>

</td>
<td>

</td>
<td>

</td>
</tr>
<tr style="background-color:#F4F4F4;">
<td>
1
</td>
<td>
y
</td>
<td>
0
</td>
<td>
1
</td>
<td>

</td>
<td>

</td>
<td>

</td>
<td>

</td>
<td>

</td>
</tr>
</table>
"""
        == minuit.init_params._repr_html_()
    )

    assert (
        """<table>
<tr style="background-color:#F4F4F4;">
<td/>
<th title="Variable name">
Name
</th>
<th title="Value of parameter">
Value
</th>
<th title="Hesse error">
Hesse Error
</th>
<th title="Minos lower error">
Minos Error-
</th>
<th title="Minos upper error">
Minos Error+
</th>
<th title="Lower limit of the parameter">
Limit-
</th>
<th title="Upper limit of the parameter">
Limit+
</th>
<th title="Is the parameter fixed in the fit">
Fixed
</th>
</tr>
<tr style="background-color:#FFFFFF;">
<td>
0
</td>
<td>
x
</td>
<td>
2
</td>
<td>
1
</td>
<td>
-1
</td>
<td>
1
</td>
<td>

</td>
<td>

</td>
<td>

</td>
</tr>
<tr style="background-color:#F4F4F4;">
<td>
1
</td>
<td>
y
</td>
<td>
1.0
</td>
<td>
0.5
</td>
<td>
-0.5
</td>
<td>
0.5
</td>
<td>

</td>
<td>

</td>
<td>

</td>
</tr>
</table>
"""
        == minuit.params._repr_html_()
    )


def test_html_params_with_limits():
    m = Minuit(
        f1,
        x=3,
        y=5,
        fix_x=True,
        error_x=0.2,
        error_y=0.1,
        limit_x=(0, None),
        limit_y=(0, 10),
        errordef=1,
        print_level=0,
    )
    assert (
        r"""<table>
<tr style="background-color:#F4F4F4;">
<td/>
<th title="Variable name">
Name
</th>
<th title="Value of parameter">
Value
</th>
<th title="Hesse error">
Hesse Error
</th>
<th title="Minos lower error">
Minos Error-
</th>
<th title="Minos upper error">
Minos Error+
</th>
<th title="Lower limit of the parameter">
Limit-
</th>
<th title="Upper limit of the parameter">
Limit+
</th>
<th title="Is the parameter fixed in the fit">
Fixed
</th>
</tr>
<tr style="background-color:#FFFFFF;">
<td>
0
</td>
<td>
x
</td>
<td>
3.0
</td>
<td>
0.2
</td>
<td>

</td>
<td>

</td>
<td>
0
</td>
<td>

</td>
<td>
yes
</td>
</tr>
<tr style="background-color:#F4F4F4;">
<td>
1
</td>
<td>
y
</td>
<td>
5.0
</td>
<td>
0.1
</td>
<td>

</td>
<td>

</td>
<td>
0
</td>
<td>
10
</td>
<td>

</td>
</tr>
</table>
"""
        == m.init_params._repr_html_()
    )


def test_html_minos(minuit):
    mes = minuit.merrors_struct
    assert (
        r"""<table>
<tr>
<th title="Parameter name">
x
</th>
<td align="center" colspan="2" style="background-color:#92CCA6;">
Valid
</td>
</tr>
<tr>
<td title="Lower and upper minos error of the parameter">
Error
</td>
<td>
-1
</td>
<td>
1
</td>
</tr>
<tr>
<td title="Validity of lower/upper minos error">
Valid
</td>
<td style="background-color:#92CCA6;">
True
</td>
<td style="background-color:#92CCA6;">
True
</td>
</tr>
<tr>
<td title="Did scan hit limit of any parameter?">
At Limit
</td>
<td style="background-color:#92CCA6;">
False
</td>
<td style="background-color:#92CCA6;">
False
</td>
</tr>
<tr>
<td title="Did scan hit function call limit?">
Max FCN
</td>
<td style="background-color:#92CCA6;">
False
</td>
<td style="background-color:#92CCA6;">
False
</td>
</tr>
<tr>
<td title="New minimum found when doing scan?">
New Min
</td>
<td style="background-color:#92CCA6;">
False
</td>
<td style="background-color:#92CCA6;">
False
</td>
</tr>
</table>

<table>
<tr>
<th title="Parameter name">
y
</th>
<td align="center" colspan="2" style="background-color:#92CCA6;">
Valid
</td>
</tr>
<tr>
<td title="Lower and upper minos error of the parameter">
Error
</td>
<td>
-0.5
</td>
<td>
0.5
</td>
</tr>
<tr>
<td title="Validity of lower/upper minos error">
Valid
</td>
<td style="background-color:#92CCA6;">
True
</td>
<td style="background-color:#92CCA6;">
True
</td>
</tr>
<tr>
<td title="Did scan hit limit of any parameter?">
At Limit
</td>
<td style="background-color:#92CCA6;">
False
</td>
<td style="background-color:#92CCA6;">
False
</td>
</tr>
<tr>
<td title="Did scan hit function call limit?">
Max FCN
</td>
<td style="background-color:#92CCA6;">
False
</td>
<td style="background-color:#92CCA6;">
False
</td>
</tr>
<tr>
<td title="New minimum found when doing scan?">
New Min
</td>
<td style="background-color:#92CCA6;">
False
</td>
<td style="background-color:#92CCA6;">
False
</td>
</tr>
</table>
"""
        == mes._repr_html_()
    )


def test_html_matrix():
    matrix = Matrix(["x", "y"], ((1.0, -0.0), (-0.0, 0.25)))
    assert (
        matrix._repr_html_()
        == r"""<table>
<tr>
<td/>

<th>
x
</th>
<th>
y
</th>
</tr>
<tr>
<th>
x
</th>
<td>
1.00
</td>
<td style="background-color:rgb(250,250,250)">
-0.00
</td>
</tr>
<tr>
<th>
y
</th>
<td style="background-color:rgb(250,250,250)">
-0.00
</td>
<td>
0.25
</td>
</tr>
</table>
"""
    )


def test_text_fmin(minuit):
    fmin = minuit.fmin
    assert (
        str(fmin)
        == r"""------------------------------------------------------------------
| FCN = 1                       |      Ncalls=34 (58 total)      |
| EDM = %.3G (Goal: 2E-07)  |            up = 1.0            |
------------------------------------------------------------------
|  Valid Min.   | Valid Param.  | Above EDM | Reached call limit |
------------------------------------------------------------------
|     True      |     True      |   False   |       False        |
------------------------------------------------------------------
| Hesse failed  |   Has cov.    | Accurate  | Pos. def. | Forced |
------------------------------------------------------------------
|     False     |     True      |   True    |   True    | False  |
------------------------------------------------------------------"""
        % fmin.edm
    )


def test_text_params(minuit):
    assert r"""------------------------------------------------------------------------------------------
|   | Name |   Value   | Hesse Err | Minos Err- | Minos Err+ | Limit-  | Limit+  | Fixed |
------------------------------------------------------------------------------------------
| 0 | x    |     0     |     1     |            |            |         |         |       |
| 1 | y    |     0     |     1     |            |            |         |         |       |
------------------------------------------------------------------------------------------""" == str(
        minuit.init_params
    )

    assert r"""------------------------------------------------------------------------------------------
|   | Name |   Value   | Hesse Err | Minos Err- | Minos Err+ | Limit-  | Limit+  | Fixed |
------------------------------------------------------------------------------------------
| 0 | x    |     2     |     1     |     -1     |     1      |         |         |       |
| 1 | y    |    1.0    |    0.5    |    -0.5    |    0.5     |         |         |       |
------------------------------------------------------------------------------------------""" == str(
        minuit.params
    )


def test_text_params_with_limits():
    m = Minuit(
        f1,
        x=3,
        y=5,
        fix_x=True,
        error_x=0.2,
        error_y=0.1,
        limit_x=(0, None),
        limit_y=(0, 10),
        errordef=1,
    )
    assert r"""------------------------------------------------------------------------------------------
|   | Name |   Value   | Hesse Err | Minos Err- | Minos Err+ | Limit-  | Limit+  | Fixed |
------------------------------------------------------------------------------------------
| 0 | x    |    3.0    |    0.2    |            |            |    0    |         |  yes  |
| 1 | y    |    5.0    |    0.1    |            |            |    0    |   10    |       |
------------------------------------------------------------------------------------------""" == str(
        m.init_params
    )


def test_text_minos(minuit):
    assert r"""-------------------------------------------------
|        x        |            Valid            |
-------------------------------------------------
|      Error      |      -1      |      1       |
|      Valid      |     True     |     True     |
|    At Limit     |    False     |    False     |
|     Max FCN     |    False     |    False     |
|     New Min     |    False     |    False     |
-------------------------------------------------
-------------------------------------------------
|        y        |            Valid            |
-------------------------------------------------
|      Error      |     -0.5     |     0.5      |
|      Valid      |     True     |     True     |
|    At Limit     |    False     |    False     |
|     Max FCN     |    False     |    False     |
|     New Min     |    False     |    False     |
-------------------------------------------------""" == str(
        minuit.minos()
    )


def test_text_matrix():
    matrix = Matrix(["x", "y"], ((1.0, -0.0), (-0.0, 0.25)))
    assert r"""-------------------
|   |     x     y |
-------------------
| x |  1.00 -0.00 |
| y | -0.00  0.25 |
-------------------""" == str(
        matrix
    )


def test_text_with_long_names():

    matrix = Matrix(["super-long-name", "x"], ((1.0, 0.1), (0.1, 1.0)))
    assert r"""-----------------------------------------------------
|                 | super-long-name               x |
-----------------------------------------------------
| super-long-name |             1.0             0.1 |
|               x |             0.1             1.0 |
-----------------------------------------------------""" == str(
        matrix
    )

    mps = Params(
        [
            Param(
                0,
                "super-long-name",
                0,
                0,
                False,
                False,
                False,
                False,
                False,
                None,
                None,
            )
        ],
        None,
    )
    assert r"""-----------------------------------------------------------------------------------------------------
|   | Name            |   Value   | Hesse Err | Minos Err- | Minos Err+ | Limit-  | Limit+  | Fixed |
-----------------------------------------------------------------------------------------------------
| 0 | super-long-name |     0     |     0     |            |            |         |         |       |
-----------------------------------------------------------------------------------------------------""" == str(
        mps
    )


def test_console_frontend_with_difficult_values():
    matrix = Matrix(("x", "y"), ((-1.23456, 0), (0, 0)))
    assert r"""---------------------
|   |      x      y |
---------------------
| x | -1.235  0.000 |
| y |  0.000  0.000 |
---------------------""" == str(
        matrix
    )

    mps = Params(
        [
            Param(
                0,
                "x",
                -1.234567e-22,
                1.234567e-11,
                True,
                False,
                False,
                False,
                False,
                None,
                None,
            )
        ],
        None,
    )

    assert r"""------------------------------------------------------------------------------------------
|   | Name |   Value   | Hesse Err | Minos Err- | Minos Err+ | Limit-  | Limit+  | Fixed |
------------------------------------------------------------------------------------------
| 0 | x    |    -0     |  1.2E-11  |            |            |         |         | CONST |
------------------------------------------------------------------------------------------""" == str(
        mps
    )



from iminuit.util import describe, make_func_code
from math import ldexp
import sys
import pytest
import platform

is_pypy = platform.python_implementation() == "PyPy"


def test_function():
    def f(x, y):
        # body is important
        a = 1
        b = 2
        return a + b

    assert describe(f, True) == ["x", "y"]


def test_class_method():
    class A:
        def f(self, x, y):
            # body is important
            a = 1
            b = 2
            return a + b

    assert describe(A().f, True) == ["x", "y"]


# unbound method
def test_class_unbound_method():
    class A:
        def f(self, x, y):
            # body is important
            a = 1
            b = 2
            return a + b

    assert describe(A.f, True) == ["self", "x", "y"]


def test_functor():
    class A:
        def __call__(self, x, y):
            # body is important
            a = 1
            b = 2
            return a + b

    assert describe(A(), True) == ["x", "y"]


@pytest.mark.skipif(
    sys.version_info >= (3, 7), reason="Not on Py 3.7; see Github issue 270"
)
def test_builtin_by_parsing_doc():
    assert describe(ldexp, True) == ["x", "i"]


def test_lambda():
    assert describe(lambda a, b: 0, True) == ["a", "b"]


def test_generic_function():
    def f(*args):
        # body is important
        a = 1
        b = 2
        return a + b

    with pytest.raises(TypeError):
        describe(f, True)


def test_generic_lambda():
    with pytest.raises(TypeError):
        describe(lambda *args: 0, True)


def test_generic_class_method():
    class A:
        def f(self, *args):
            # body is important
            a = 1
            b = 2
            return a + b

    with pytest.raises(TypeError):
        describe(A().f, True)


def test_generic_functor():
    class A:
        def __call__(self, *args):
            # body is important
            a = 1
            b = 2
            return a + b

    with pytest.raises(TypeError):
        describe(A(), True)


def test_generic_functor_with_fake_func():
    class A:
        def __init__(self):
            self.func_code = make_func_code(["x", "y"])

        def __call__(self, *args):
            # body is important
            a = 1
            b = 2
            return a + b

    assert describe(A(), True) == ["x", "y"]
